/******************************************************************************
 * Copyright 2020 The siasun Transport Authors. All Rights Reserved.
 * Author: madesheng
 * Licensed under the Apache License, Version 1.0 (the "License");
 *****************************************************************************/
/**
 * @file
 */

#ifndef ROBOT_ROBOT_LOCALIZATIONDATA_H_
#define ROBOT_ROBOT_LOCALIZATIONDATA_H_

#include <math.h>
#include <string>
#include <iomanip>
#include "../common/gloghelper.h"

/**
 * @namespace robot::localization
 * @brief robot::localization
 */
namespace robot {
namespace localizationspace {

#define ROBOT_SUCCESS     0
#define ROBOT_FAILTURE   -1


const double RADIANS_PER_DEGREE = M_PI/180.0;
const double DEGREES_PER_RADIAN = 180.0/M_PI;

// WGS84 Parameters
const double WGS84_A = 6378137.0;		// major axis
const double WGS84_B = 6356752.31424518;	// minor axis
const double WGS84_F = 0.0033528107;		// ellipsoid flattening
const double WGS84_E = 0.0818191908;		// first eccentricity
const double WGS84_EP = 0.0820944379;		// second eccentricity

// UTM Parameters
const double UTM_K0 = 0.9996;			// scale factor
const double UTM_FE = 500000.0;		// false easting
const double UTM_FN_N = 0.0;			// false northing on north hemisphere
const double UTM_FN_S = 10000000.0;		// false northing on south hemisphere
const double UTM_E2 = (WGS84_E*WGS84_E);	// e^2
const double UTM_E4 = (UTM_E2*UTM_E2);		// e^4
const double UTM_E6 = (UTM_E4*UTM_E2);		// e^6
const double UTM_EP2 = (UTM_E2/(1-UTM_E2));	// e'^2

/**GPRMC  <7> 地面速率 单位为节(knots)，1节=1.852km/h**/
//#define SPEED_FACTOR   1.852
/**IFDYN  <4> 水平合成速度 单位为 m/s 1m/s=3.6km/h**/
#define SPEED_FACTOR   3.6
/**
* NMEA0183协议-GPRMC
*/
struct GPRMC
{
    // <1> UTC时间，hhmmss.ss时分秒格式
    double utcTime;
    // <2> 定位状态，A=有效定位，V=无效定位
    char locationStatus[2];
    // <3> 纬度ddmm.mmmmmmmm（度分）格式（前面的0也将被传输）
    double Latitude;
    // <4> 纬度半球N（北半球）或S（南半球）
    char LatType[2];
    // <5> 经度dddmm.mmmmmmmm（度分）格式（前面的0也将被传输）
    double Longitude;
    // <6> 经度半球E（东经）或W（西经）
    char LonType[2];
    // <7> 地面速率（000.0~999.9节，前面的0也将被传输）单位为节(knots)，1节=1.852km/h
    double Speed;
    // <8> 地面航向（000.0~359.9度，以真北为参考基准，前面的0也将被传输）
    double Deviated;
    // <9> UTC日期，ddmmyy（日月年）格式
    char Date[5];
    // <10> 磁偏角（000.0~180.0度，前面的0也将被传输）
    //double Magnetic;
    // <11> 磁偏角方向，E（东）或W（西）
    //char MagneticDirection[2];
    // <12> 模式指示（仅NMEA0183 3.00版本输出，A=自主定位，D=差分，E=估算，N=数据无效）
    //char Mode[2];
    void print()
    {
        LOG(INFO)<< "utcTime:"<<utcTime;
        LOG(INFO)<< "locationStatus:"<<std::string(locationStatus);
        LOG(INFO)<< "Latitude:"<<std::setprecision(18)<<Latitude;
        LOG(INFO)<< "LatType:"<<std::string(LatType);
        LOG(INFO)<< "Longitude:"<<std::setprecision(18)<<Longitude;
        LOG(INFO)<< "LonType:"<<std::string(LonType);
        LOG(INFO)<< "Speed:"<<Speed;
        LOG(INFO)<< "Deviated:"<<Deviated;
        LOG(INFO)<< "Date:"<<std::string(Date);
//        LOG(INFO)<< "Magnetic:"<<Magnetic;
//        LOG(INFO)<< "MagneticDirection:"<<std::string(MagneticDirection);
//        LOG(INFO)<< "Mode:"<<std::string(Mode);
    }
};

/**
* NMEA0183协议-GPGGA
*/
struct GPGGA
{
    // 字段1：UTC 时间，hhmmss.sss，时分秒格式
    double utcTime;
    // 字段2：纬度ddmm.mmmmmmmm，度分格式（前导位数不足则补0）
    double Latitude;
    // 字段3：纬度N（北纬）或S（南纬）
    char LatType[2];
    // 字段4：经度dddmm.mmmmmmmm，度分格式（前导位数不足则补0）
    double Longitude;
    // 字段5：经度E（东经）或W（西经）
    char LonType[2];
    // 字段6：GPS状态，0=未定位，1=非差分定位，2=差分定位，3=无效PPS，6=正在估算
    char GPSStatus[2];
    // 字段7：正在使用的卫星数量（00 - 12）（前导位数不足则补0）
    int SatelliteNumber;
    // 字段8：HDOP水平精度因子HDOP=0.0~9.9
    double Hdop;
    // 字段9：海拔高度（-9999.9 - 99999.9）
    double Height;
    // 字段10：海拔高度单位（米）
    char  HeightUint[2];
    void print()
    {
        LOG(INFO)<< "utcTime:"<<utcTime;
        LOG(INFO)<< "Latitude:"<<std::setprecision(18)<<Latitude;
        LOG(INFO)<< "LatType:"<<std::string(LatType);
        LOG(INFO)<< "Longitude:"<<std::setprecision(18)<<Longitude;
        LOG(INFO)<< "LonType:"<<std::string(LonType);
        LOG(INFO)<< "GPSStatus:"<<std::string(GPSStatus);
        LOG(INFO)<< "SatelliteNumber:"<<SatelliteNumber;
        LOG(INFO)<< "Hdop:"<<Hdop;
        LOG(INFO)<< "Height:"<<Height;
        LOG(INFO)<< "HeightUint:"<<std::string(HeightUint);
    }
};

/**
* IFDYN
*/
struct IFDYN
{
  // 字段1：车辆航向角 -180~180 度 0.001
  double YawAngle;
  // 字段2：车辆横滚角，-180~180 度 0.001
  double RollAngle;
  // 字段3：车辆俯仰角 -90~90 度 0.001
  double PitchAngle;
  // 字段4：水平合成速度-地球坐标系中的水平速度 0~200 m/s 0.001
  double VelHori;
  // 字段5：东向水平速度-地球坐标系中的东向水平速度分量 -190~190 m/s 0.001
  double VelEast;
  // 字段6：北向水平速度-地球坐标系中的北向水平速度分量 -190~190 m/s 0.001
  double VelNorth;
  // 字段7：地球坐标系中的地向速度 -190~190 m/s 0.001
  double VelD;
  // 字段8：纵向加速度-车辆坐标系x轴加速度  -58.5~58.8 m/s2  0.001
  double AccelX;
  // 字段9：横向加速度-车辆坐标系y轴加速度 -58.5~58.8 m/s2   0.001
  double AccelY;
  // 字段10：地向加速度-车辆坐标系z轴加速度 -58.5~58.8 m/s2  0.001
  double  AccelZ;
  // 字段11：横滚角速度-车辆坐标系x轴角速度 -300~300 。/s    0.001
  double  AngleRateX;
  // 字段12：俯仰角速度-车辆坐标系y轴角速度 -300~300 。/s    0.001
  double  AngleRateY;
  // 字段13：航向加速度-车辆坐标系z轴角速度 -300~300 。/s    0.001
  double  AngleRateZ;
  void print()
  {
      LOG(INFO)<< "YawAngle:"<<std::setprecision(5)<<YawAngle;
      LOG(INFO)<< "RollAngle:"<<std::setprecision(4)<<RollAngle;
      LOG(INFO)<< "PitchAngle:"<<std::setprecision(4)<<PitchAngle;
      LOG(INFO)<< "VelHori:"<<std::setprecision(4)<<VelHori;
      LOG(INFO)<< "VelEast:"<<std::setprecision(4)<<VelEast;
      LOG(INFO)<< "VelNorth:"<<std::setprecision(4)<<VelNorth;
      LOG(INFO)<< "VelD:"<<std::setprecision(4)<<VelD;
      LOG(INFO)<< "AccelX:"<<std::setprecision(4)<<AccelX;
      LOG(INFO)<< "AccelY:"<<std::setprecision(4)<<AccelY;
      LOG(INFO)<< "AccelZ:"<<std::setprecision(4)<<AccelZ;
      LOG(INFO)<< "AngleRateX:"<<std::setprecision(4)<<AngleRateX;
      LOG(INFO)<< "AngleRateY:"<<std::setprecision(4)<<AngleRateY;
      LOG(INFO)<< "AngleRateZ:"<<std::setprecision(4)<<AngleRateZ;
  }

};

/**
* GPFPD
*/
struct GPFPD
{
  // <1> GPS周-自1980-1-6至当前的星期数（格林尼治时间）
  unsigned int GPSWeek;
  // <2> GPS周秒-自本周日0:00:00至当前的秒数（格林尼治时间）
  char GPSWeekSecond[16];
  // <3> 航向角 0~360度
  double Heading;
  // <4> 俯仰角 -90~90度
  double Pitch;
  // <5> 车辆横滚角 -180~180度
  double Roll;
  // <6> 纬度 -90~90度
  double Latitude;
  // <7> 经度  -180~180度
  double Longitude;
  // <8> 高度 单位米
  double Alt;
  // <9> 东向水平速度-地球坐标系中的东向水平速度分量 -190~190 m/s 0.001
  double VelEast;
  // <10> 北向水平速度-地球坐标系中的北向水平速度分量 -190~190 m/s 0.001
  double VelNorth;
  // <11> 天向速度-地球坐标系中的天向速度 -190~190 m/s 0.001
  double VelU;
  // <12> 基线长度 单位米
  double BaseLine;
  // <13> 天线1卫星数
  int Sat1Num;
  // <14> 天线2卫星数
  int Sat2Num;
  // <15> 状态  定位状态 00：定位无效   03：GNSS单点定位   04：组合惯导单点定位  05：GNSS差分定位
  //06：组合惯导轮速辅助  08：纯惯导推算  4B：组合惯导RTK固定  5B：组合惯导RTK浮点
  char Status[4];
  void print()
  {
      LOG(INFO)<< "GPSWeek:"<<GPSWeek;
      LOG(INFO)<< "GPSWeekSecond:"<<std::string(GPSWeekSecond);
      LOG(INFO)<< "Heading:"<<std::setprecision(5)<<Heading;
      LOG(INFO)<< "Pitch:"<<std::setprecision(3)<<Pitch;
      LOG(INFO)<< "Roll:"<<std::setprecision(3)<<Roll;
      LOG(INFO)<< "Latitude:"<<std::setprecision(9)<<Latitude;
      LOG(INFO)<< "Longitude:"<<std::setprecision(10)<<Longitude;
      LOG(INFO)<< "Alt:"<<std::setprecision(3)<<Alt;
      LOG(INFO)<< "VelEast:"<<std::setprecision(4)<<VelEast;
      LOG(INFO)<< "VelNorth:"<<std::setprecision(4)<<VelNorth;
      LOG(INFO)<< "VelU:"<<std::setprecision(4)<<VelU;
      LOG(INFO)<< "BaseLine:"<<std::setprecision(4)<<BaseLine;
      LOG(INFO)<< "Sat1Num:"<<Sat1Num;
      LOG(INFO)<< "Sat2Num:"<<Sat2Num;
      LOG(INFO)<< "Status:"<<std::string(Status);

//      printf("*************GPFDP_data*****************\n");
//        printf("GPSWeek:%d\n",gpfdp_data->GPSWeek);
//          printf("GPSTime:%s\n",gpfdp_data->GPSTime);
//          printf("Heading:%f\n",gpfdp_data->Heading);
//          printf("Pitch:%f\n",gpfdp_data->Pitch);
//          printf("Roll:%f\n",gpfdp_data->Roll);
//          printf("Latitude:%12.9lf\n",gpfdp_data->Latitude);
//          printf("Longitude:%12.9lf\n",gpfdp_data->Longitude);
//          printf("Altitude:%12.9lf\n",gpfdp_data->Altitude);
//          printf("[Ve Vn Vu]=[%f %f %f]\n",gpfdp_data->Ve,gpfdp_data->Vn,gpfdp_data->Vu);
//        printf("Baseline =%f\n",gpfdp_data->Baseline);
//          printf("Status:%s\n",gpfdp_data->Status);
  }

};

/**
* NMEA0183协议
*/
struct NMEA0183
{
    struct GPRMC gprmc;
    struct GPGGA gpgga;
    struct IFDYN ifdyn;
    struct GPFPD gpfpd;
};



} // localization
} // robot

#endif //ROBOT_ROBOT_LOCALIZATIONDATA_H_
